@@ -3,24 +3,23 @@
 import ctypes
 import platform
 import inspect
-import pytz
 from ctypes import *
-from time import tzname
+import pytz
 
 try:
     from typing import Any
-except:
+except BaseException:
     pass
 
 from .error import *
-from .bind import *
-from .field import *
 from .schemaless import *
 
 _UNSUPPORTED = {}
 
 
 # C interface class
+
+
 class TaosOption:
     Locale = 0
     Charset = 1
@@ -59,13 +58,50 @@ def _load_taos():
 
 _libtaos = _load_taos()
 
+_libtaos.taos_get_client_info.restype = c_char_p
+
+
+def taos_get_client_info():
+    # type: () -> str
+    """Get client version info."""
+    return _libtaos.taos_get_client_info().decode("utf-8")
+
+
+IS_V3 = False
+
+if taos_get_client_info().split(".")[0] < "3":
+    from .field import CONVERT_FUNC, CONVERT_FUNC_BLOCK, TaosFields, TaosField, set_tz
+else:
+    from .field import CONVERT_FUNC, CONVERT_FUNC_BLOCK, TaosFields, TaosField, set_tz
+
+    # use _v3s TaosField overwrite _v2s here, dont change import order
+    from .field_v3 import CONVERT_FUNC_BLOCK_v3, TaosFields, TaosField
+    from .constants import FieldType
+
+    IS_V3 = True
+
 _libtaos.taos_fetch_fields.restype = ctypes.POINTER(TaosField)
-_libtaos.taos_init.restype = None
+
+try:
+    _libtaos.taos_init.restype = None
+except Exception as err:
+    _UNSUPPORTED["taos_init"] = err
+
 _libtaos.taos_connect.restype = ctypes.c_void_p
 _libtaos.taos_fetch_row.restype = ctypes.POINTER(ctypes.c_void_p)
 _libtaos.taos_errstr.restype = ctypes.c_char_p
-_libtaos.taos_subscribe.restype = ctypes.c_void_p
-_libtaos.taos_consume.restype = ctypes.c_void_p
+
+try:
+    _libtaos.taos_subscribe.restype = ctypes.c_void_p
+except Exception as err:
+    _UNSUPPORTED["taos_subscribe"] = err
+
+try:
+    _libtaos.taos_consume.restype = ctypes.c_void_p
+    _libtaos.taos_consume.argstype = (c_void_p,)
+except Exception as err:
+    _UNSUPPORTED["taos_consume"] = err
+
 _libtaos.taos_fetch_lengths.restype = ctypes.POINTER(ctypes.c_int)
 _libtaos.taos_free_result.restype = None
 _libtaos.taos_query.restype = ctypes.POINTER(ctypes.c_void_p)
@@ -77,7 +113,6 @@ except AttributeError:
 finally:
     None
 
-
 _libtaos.taos_options.restype = None
 _libtaos.taos_options.argtypes = (c_int, c_void_p)
 
@@ -105,15 +140,6 @@ def taos_cleanup():
     _libtaos.taos_cleanup()
 
 
-_libtaos.taos_get_client_info.restype = c_char_p
-
-
-def taos_get_client_info():
-    # type: () -> str
-    """Get client version info."""
-    return _libtaos.taos_get_client_info().decode("utf-8")
-
-
 _libtaos.taos_get_server_info.restype = c_char_p
 _libtaos.taos_get_server_info.argtypes = (c_void_p,)
 
@@ -183,7 +209,10 @@ def taos_connect(host=None, user="root", password="taosdata", db=None, port=0):
     connection = cast(_libtaos.taos_connect(_host, _user, _password, _db, _port), c_void_p)
 
     if connection.value is None:
-        raise ConnectionError("connect to TDengine failed")
+        null_ptr = c_void_p(None)
+        errno = taos_errno(null_ptr)
+        errstr = taos_errstr(null_ptr)
+        raise ConnectionError(errstr, errno)
     return connection
 
 
@@ -239,7 +268,10 @@ def taos_connect_auth(host=None, user="root", auth="", db=None, port=0):
     connection = c_void_p(_libtaos.taos_connect_auth(_host, _user, _auth, _db, _port))
 
     if connection.value is None:
-        raise ConnectionError("connect to TDengine failed")
+        null_ptr = c_void_p(None)
+        errno = taos_errno(null_ptr)
+        errstr = taos_errstr(null_ptr)
+        raise ConnectionError(errstr, errno)
     return connection
 
 
@@ -296,7 +328,8 @@ def taos_affected_rows(result):
 
 
 subscribe_callback_type = CFUNCTYPE(None, c_void_p, c_void_p, c_void_p, c_int)
-_libtaos.taos_subscribe.restype = c_void_p
+
+
 # _libtaos.taos_subscribe.argtypes = c_void_p, c_int, c_char_p, c_char_p, subscribe_callback_type, c_void_p, c_int
 
 
@@ -307,7 +340,7 @@ def taos_subscribe(connection, restart, topic, sql, interval, callback=None, par
     @sql string, sql statement for data query, must be a 'select' statement.
     @topic string, name of this subscription
     """
-    if callback != None:
+    if callback is not None:
         callback = subscribe_callback_type(callback)
     return c_void_p(
         _libtaos.taos_subscribe(
@@ -322,17 +355,16 @@ def taos_subscribe(connection, restart, topic, sql, interval, callback=None, par
     )
 
 
-_libtaos.taos_consume.restype = c_void_p
-_libtaos.taos_consume.argstype = (c_void_p,)
-
-
 def taos_consume(sub):
     """Consume data of a subscription"""
     return c_void_p(_libtaos.taos_consume(sub))
 
 
-_libtaos.taos_unsubscribe.restype = None
-_libtaos.taos_unsubscribe.argstype = c_void_p, c_int
+try:
+    _libtaos.taos_unsubscribe.restype = None
+    _libtaos.taos_unsubscribe.argstype = c_void_p, c_int
+except Exception as err:
+    _UNSUPPORTED["taos_unsubscribe"] = err
 
 
 def taos_unsubscribe(sub, keep_progress):
@@ -356,6 +388,14 @@ def taos_use_result(result):
     return fields
 
 
+_libtaos.taos_is_null.restype = c_bool
+_libtaos.taos_is_null.argtypes = ctypes.c_void_p, c_int, c_int
+
+
+def taos_is_null(result, row, col):
+    return _libtaos.taos_is_null(result, row, col)
+
+
 _libtaos.taos_fetch_block.restype = c_int
 _libtaos.taos_fetch_block.argtypes = c_void_p, c_void_p
 
@@ -368,7 +408,49 @@ def taos_fetch_block_raw(result):
     return pblock, abs(num_of_rows)
 
 
-def taos_fetch_block(result, fields=None, field_count=None):
+if not IS_V3:
+    pass
+else:
+    _libtaos.taos_get_column_data_offset.restype = ctypes.POINTER(ctypes.c_int)
+    _libtaos.taos_get_column_data_offset.argtypes = ctypes.c_void_p, c_int
+
+
+def taos_get_column_data_offset(result, field, rows):
+    # Make sure to call taos_get_column_data_offset after taos_fetch_block()
+    offsets = _libtaos.taos_get_column_data_offset(result, field)
+    if not offsets:
+        raise OperationalError("offsets empty, use taos_fetch_block before it")
+    return offsets[:rows]
+
+
+def taos_fetch_block_v3(result, fields=None, field_count=None):
+    if fields is None:
+        fields = taos_fetch_fields(result)
+    if field_count is None:
+        field_count = taos_field_count(result)
+    pblock = ctypes.c_void_p(0)
+    num_of_rows = _libtaos.taos_fetch_block(result, ctypes.byref(pblock))
+    if num_of_rows == 0:
+        return None, 0
+    precision = taos_result_precision(result)
+    blocks = [None] * field_count
+    for i in range(len(fields)):
+        data = ctypes.cast(pblock, ctypes.POINTER(ctypes.c_void_p))[i]
+        if fields[i]["type"] not in CONVERT_FUNC_BLOCK_v3 and fields[i]["type"] not in CONVERT_FUNC_BLOCK:
+            raise DatabaseError("Invalid data type returned from database")
+        offsets = []
+        is_null = []
+        if fields[i]["type"] in (FieldType.C_VARCHAR, FieldType.C_NCHAR, FieldType.C_JSON):
+            offsets = taos_get_column_data_offset(result, i, num_of_rows)
+            blocks[i] = CONVERT_FUNC_BLOCK_v3[fields[i]["type"]](data, is_null, num_of_rows, offsets, precision)
+        else:
+            is_null = [taos_is_null(result, j, i) for j in range(num_of_rows)]
+            blocks[i] = CONVERT_FUNC_BLOCK[fields[i]["type"]](data, is_null, num_of_rows, offsets, precision)
+
+    return blocks, abs(num_of_rows)
+
+
+def taos_fetch_block_v2(result, fields=None, field_count=None):
     pblock = ctypes.c_void_p(0)
     num_of_rows = _libtaos.taos_fetch_block(result, ctypes.byref(pblock))
     if num_of_rows == 0:
@@ -384,11 +466,17 @@ def taos_fetch_block(result, fields=None, field_count=None):
         data = ctypes.cast(pblock, ctypes.POINTER(ctypes.c_void_p))[i]
         if fields[i]["type"] not in CONVERT_FUNC:
             raise DatabaseError("Invalid data type returned from database")
-        blocks[i] = CONVERT_FUNC_BLOCK[fields[i]["type"]](data, num_of_rows, fieldLen[i], precision)
+        is_null = [taos_is_null(result, j, i) for j in range(num_of_rows)]
+        blocks[i] = CONVERT_FUNC_BLOCK[fields[i]["type"]](data, is_null, num_of_rows, fieldLen[i], precision)
 
     return blocks, abs(num_of_rows)
 
 
+if not IS_V3:
+    taos_fetch_block = taos_fetch_block_v2
+else:
+    taos_fetch_block = taos_fetch_block_v3
+
 _libtaos.taos_fetch_row.restype = c_void_p
 _libtaos.taos_fetch_row.argtypes = (c_void_p,)
 
@@ -418,7 +506,7 @@ def taos_fetch_row(result, fields):
             if data is None:
                 blocks[i] = [None]
             else:
-                blocks[i] = CONVERT_FUNC[fields[i].type](data, num_of_rows, field_lens[i], precision)
+                blocks[i] = CONVERT_FUNC[fields[i].type](data, [False], num_of_rows, field_lens[i], precision)
     else:
         return None, 0
     return blocks, abs(num_of_rows)
@@ -429,7 +517,7 @@ _libtaos.taos_free_result.argtypes = (c_void_p,)
 
 def taos_free_result(result):
     # type: (c_void_p) -> None
-    if result != None:
+    if result is not None:
         _libtaos.taos_free_result(result)
 
 
@@ -666,8 +754,11 @@ def taos_stmt_set_tbname_tags(stmt, name, tags):
         raise StatementError(msg=taos_stmt_errstr(stmt), errno=res)
 
 
-_libtaos.taos_stmt_is_insert.restype = c_int
-_libtaos.taos_stmt_is_insert.argstype = (c_void_p, POINTER(c_int))
+try:
+    _libtaos.taos_stmt_is_insert.restype = c_int
+    _libtaos.taos_stmt_is_insert.argstype = (c_void_p, POINTER(c_int))
+except Exception as err:
+    _UNSUPPORTED["taos_stmt_is_insert"] = err
 
 
 def taos_stmt_is_insert(stmt):
@@ -682,8 +773,11 @@ def taos_stmt_is_insert(stmt):
     return is_insert == 0
 
 
-_libtaos.taos_stmt_num_params.restype = c_int
-_libtaos.taos_stmt_num_params.argstype = (c_void_p, POINTER(c_int))
+try:
+    _libtaos.taos_stmt_num_params.restype = c_int
+    _libtaos.taos_stmt_num_params.argstype = (c_void_p, POINTER(c_int))
+except Exception as err:
+    _UNSUPPORTED["taos_stmt_num_params"] = err
 
 
 def taos_stmt_num_params(stmt):
@@ -698,8 +792,11 @@ def taos_stmt_num_params(stmt):
     return num_params.value
 
 
-_libtaos.taos_stmt_bind_param.restype = c_int
-_libtaos.taos_stmt_bind_param.argstype = (c_void_p, c_void_p)
+try:
+    _libtaos.taos_stmt_bind_param.restype = c_int
+    _libtaos.taos_stmt_bind_param.argstype = (c_void_p, c_void_p)
+except Exception as err:
+    _UNSUPPORTED["taos_stmt_bind_param"] = err
 
 
 def taos_stmt_bind_param(stmt, bind):
@@ -756,8 +853,11 @@ def taos_stmt_bind_single_param_batch(stmt, bind, col):
         raise StatementError(msg=taos_stmt_errstr(stmt), errno=res)
 
 
-_libtaos.taos_stmt_add_batch.restype = c_int
-_libtaos.taos_stmt_add_batch.argstype = (c_void_p,)
+try:
+    _libtaos.taos_stmt_add_batch.restype = c_int
+    _libtaos.taos_stmt_add_batch.argstype = (c_void_p,)
+except Exception as err:
+    _UNSUPPORTED["taos_stmt_add_batch"] = err
 
 
 def taos_stmt_add_batch(stmt):
@@ -770,8 +870,11 @@ def taos_stmt_add_batch(stmt):
         raise StatementError(msg=taos_stmt_errstr(stmt), errno=res)
 
 
-_libtaos.taos_stmt_execute.restype = c_int
-_libtaos.taos_stmt_execute.argstype = (c_void_p,)
+try:
+    _libtaos.taos_stmt_execute.restype = c_int
+    _libtaos.taos_stmt_execute.argstype = (c_void_p,)
+except Exception as err:
+    _UNSUPPORTED["taos_stmt_execute"] = err
 
 
 def taos_stmt_execute(stmt):
@@ -844,9 +947,253 @@ def taos_schemaless_insert(connection, lines, protocol, precision):
     taos_free_result(res)
     return affected_rows
 
+try:
+    _libtaos.tmq_conf_new.restype = c_void_p
+except Exception as err:
+    _UNSUPPORTED["tmq_conf_new"] = err
+    
+def tmq_conf_new():
+    # type: () -> c_void_p
+    _check_if_supported()
+    return c_void_p(_libtaos.tmq_conf_new())
 
-def _check_if_supported():
-    func = inspect.stack()[1][3]
+try:
+    _libtaos.tmq_conf_set.restype = c_int
+    _libtaos.tmq_conf_set.argtypes = (c_void_p, c_char_p, c_char_p)
+except Exception as err:
+    _UNSUPPORTED["tmq_conf_set"] = err 
+
+def tmq_conf_set(conf, key, value):
+    # type: (c_void_p, c_char_p, c_char_p) -> None
+    _check_if_supported()
+    res = _libtaos.tmq_conf_set(conf, ctypes.c_char_p(key.encode("utf-8")), ctypes.c_char_p(value.encode("utf-8")))
+    if res != 0:
+        raise TmqError(msg=f"fail to execute tmq_conf_set({key},{value}), code={res}", errno=res)
+
+try:
+    _libtaos.tmq_conf_destroy.argtypes = (c_void_p,)
+    _libtaos.tmq_conf_destroy.restype = None
+except Exception as err:
+    _UNSUPPORTED["tmq_conf_destroy"] = err
+
+def tmq_conf_destroy(conf):
+    # type (c_void_p,) -> None
+    _check_if_supported()
+    _libtaos.tmq_conf_destroy(conf)
+
+tmq_commit_cb = CFUNCTYPE(c_void_p, c_int, c_void_p, c_void_p)
+try:
+    _libtaos.tmq_conf_set_auto_commit_cb.argtypes = (c_void_p, tmq_commit_cb, c_void_p)
+    _libtaos.tmq_conf_set_auto_commit_cb.restype = None
+except Exception as err:
+    _UNSUPPORTED["tmq_conf_set_auto_commit_cb"] = err
+    
+def tmq_conf_set_auto_commit_cb(conf, cb, param):
+    # type (c_void_p, tmq_commit_cb, c_void_p) -> None
+    _check_if_supported()
+    _libtaos.tmq_conf_set_auto_commit_cb(conf, tmq_commit_cb(cb), param)
+
+try:    
+    _libtaos.tmq_consumer_new.restype = c_void_p
+    _libtaos.tmq_consumer_new.argtypes = (c_void_p, c_char_p, c_int)
+except Exception as err:
+    _UNSUPPORTED["tmq_consumer_new"] = err
+
+def tmq_consumer_new(conf, errstrlen=0):
+    # type (c_void_p, c_char_p, c_int) -> c_void_p
+    _check_if_supported()
+    buf = ctypes.create_string_buffer(errstrlen)
+    tmq = cast(_libtaos.tmq_consumer_new(conf, buf, errstrlen), c_void_p)
+    if tmq.value is None:
+        raise TmqError("failed on tmq_consumer_new")
+    return tmq
+
+try:
+    _libtaos.tmq_list_new.restype = c_void_p
+except Exception as err:
+    _UNSUPPORTED["tmq_list_new"] = err
+
+def tmq_list_new():
+    # type () -> c_void_p
+    _check_if_supported()
+    return c_void_p(_libtaos.tmq_list_new())
+
+try:
+    _libtaos.tmq_list_append.restype = c_int
+    _libtaos.tmq_list_append.argtypes = (c_void_p, c_char_p)
+except Exception as err:
+    _UNSUPPORTED["tmq_list_append"] = err
+
+def tmq_list_append(list, topic):
+    # type (c_void_p, c_char_p) -> None
+    _check_if_supported()
+    res = _libtaos.tmq_list_append(list, ctypes.c_char_p(topic.encode("utf-8")))
+    return res
+
+try:
+    _libtaos.tmq_list_destroy.restype = None
+    _libtaos.tmq_list_destroy.argtypes = (c_void_p,)
+except Exception as err:
+    _UNSUPPORTED["tmq_list_destroy"] = err
+    
+def tmq_list_destroy(list):
+    _check_if_supported()
+    # type (c_void_p,) -> None
+    _libtaos.tmq_list_destroy(list)
+
+try:    
+    _libtaos.tmq_list_get_size.restype = c_int
+    _libtaos.tmq_list_get_size.argtypes = (c_void_p,)
+except Exception as err:
+    _UNSUPPORTED["tmq_list_get_size"] = err
+
+try:
+    _libtaos.tmq_list_to_c_array.restype = ctypes.POINTER(ctypes.c_char_p)
+    _libtaos.tmq_list_to_c_array.argtypes = (c_void_p,)
+except Exception as err:
+    _UNSUPPORTED["tmq_list_to_c_array"] = err
+
+def tmq_list_to_c_array(list):
+    # type (c_void_p,) -> [string]
+    _check_if_supported()
+    _check_if_supported("tmq_list_get_size")
+    c_array = _libtaos.tmq_list_to_c_array(list)
+    size = _libtaos.tmq_list_get_size(list)
+    res = []
+    for i in range(size):
+        res.append(c_array[i].decode("utf-8"))
+    return res
+
+try:
+    _libtaos.tmq_subscribe.argtypes = (c_void_p, c_void_p)
+    _libtaos.tmq_subscribe.restype = c_int
+except Exception as err:
+    _UNSUPPORTED["tmq_subscribe"] = err
+
+def tmq_subscribe(tmq, list):
+    # type (c_void_p, c_void_p) -> None
+    _check_if_supported()
+    res =  _libtaos.tmq_subscribe(tmq, list)
+    if res != 0:
+        raise TmqError(msg="failed on tmq_subscribe()", errno=res)
+
+try:
+    _libtaos.tmq_unsubscribe.argtypes = (c_void_p,)
+    _libtaos.tmq_unsubscribe.restype = c_int
+except Exception as err:
+    _UNSUPPORTED["tmq_unsubscribe"] = err
+
+def tmq_unsubscribe(tmq):
+    # type (c_void_p,) -> None
+    _check_if_supported()
+    res = _libtaos.tmq_unsubscribe(tmq)
+    if res != 0:
+        raise TmqError(msg="failed on tmq_unsubscribe()", errno=res)
+
+try:
+    _libtaos.tmq_subscription.argtypes = (c_void_p, c_void_p)
+    _libtaos.tmq_subscription.restype = c_int
+except Exception as err:
+    _UNSUPPORTED["tmq_subscription"] = err
+
+def tmq_subscription(tmq):
+    # type (c_void_p, c_void_p) -> [string]
+    _check_if_supported()
+    topics = tmq_list_new()
+    res =  _libtaos.tmq_subscription(tmq, byref(topics))
+    if res != 0:
+        raise TmqError(msg="failed on tmq_subscription()", errno=res)
+    res =  tmq_list_to_c_array(topics)
+    tmq_list_destroy(topics)
+    return res;
+    
+try:
+    _libtaos.tmq_consumer_poll.argtypes = (c_void_p, c_int64)
+    _libtaos.tmq_consumer_poll.restype = c_void_p
+except Exception as err:
+    _UNSUPPORTED["tmq_consumer_poll"] = err
+
+def tmq_consumer_poll(tmq, wait_time):
+    # type (c_void_p, c_int64) -> c_void_p
+    _check_if_supported()
+    return c_void_p(_libtaos.tmq_consumer_poll(tmq, wait_time))
+
+try:
+    _libtaos.tmq_consumer_close.argtypes = (c_void_p,)
+    _libtaos.tmq_consumer_close.restype = c_int
+except Exception as err:
+    _UNSUPPORTED["tmq_consumer_close"] = err
+
+def tmq_consumer_close(tmq):
+    # type (c_void_p,) -> None
+    _check_if_supported()
+    res =  _libtaos.tmq_consumer_close(tmq)
+    if res != 0:
+        raise TmqError(msg="failed on tmq_consumer_close()", errno=res)
+try:
+    _libtaos.tmq_commit.argtypes = (c_void_p, c_void_p, c_int32)
+    _libtaos.tmq_commit.restype = c_int
+except Exception as err:
+    _UNSUPPORTED["tmq_commit"] = err
+
+def tmq_commit(tmq, offset, _async):
+    # type (c_void_p, c_void_p, c_int32) -> None
+    _check_if_supported()
+    res = _libtaos.tmq_commit(tmq, offset, _async)
+    if res != 0:
+        raise TmqError(msg="failed on tmq_commit()", errno=res)
+
+try:
+    _libtaos.tmq_get_topic_name.argtypes = (c_void_p,)
+    _libtaos.tmq_get_topic_name.restype = c_char_p
+except Exception as err:
+    _UNSUPPORTED["tmq_get_topic_name"] = err
+
+def tmq_get_topic_name(res):
+    # type (c_void_p,) -> string
+    _check_if_supported()
+    return _libtaos.tmq_get_topic_name(res).decode("utf-8")
+
+try:
+    _libtaos.tmq_get_vgroup_id.argtypes = (c_void_p,)
+    _libtaos.tmq_get_vgroup_id.restype = c_int
+except Exception as err:
+    _UNSUPPORTED["tmq_get_vgroup_id"] = err
+
+def tmq_get_vgroup_id(res):
+    # type (c_void_p,) -> int
+    _check_if_supported()
+    return _libtaos.tmq_get_vgroup_id(res)
+
+try:
+    _libtaos.tmq_get_table_name.argtypes = (c_void_p,)
+    _libtaos.tmq_get_table_name.restype = c_char_p
+except Exception as err:
+    _UNSUPPORTED["tmq_get_table_name"] = err
+
+def tmq_get_table_name(res):
+    # type (c_void_p,) -> string
+    _check_if_supported()
+    tb_name = _libtaos.tmq_get_table_name(res)
+    if tb_name:
+        return tb_name.decode("utf-8")
+    else:
+        print("change msg.with.table.name to true in tmq config")
+
+try:        
+    _libtaos.tmq_get_db_name.argtypes = (c_void_p,)
+    _libtaos.tmq_get_db_name.restype = c_char_p
+except Exception as err:
+    _UNSUPPORTED["tmq_get_db_name"] = err
+
+def tmq_get_db_name(res):
+    # type (c_void_p,) -> string
+    _check_if_supported()
+    return _libtaos.tmq_get_db_name(res).decode("utf-8")
+
+def _check_if_supported(func=None):
+    if not func:
+        func = inspect.stack()[1][3]
     if func in _UNSUPPORTED:
         raise InterfaceError(
             "C function %s is not supported in v%s: %s" % (func, taos_get_client_info(), _UNSUPPORTED[func])
